import logging
import re
import sys

SERVER_LOG = logging.getLogger('app.server')


class Port:
    def __set__(self, instance, value):
        if not 1023 < value < 65536:
            SERVER_LOG.critical(
                f'Попытка запуска сервера с указанием неподходящего порта '
                f'{value}. Допустимы адреса с 1024 до 65535.')
            sys.exit(1)
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        return instance.__dict__[self.name]


class Host:
    def __set__(self, instance, value):
        regex = "^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
        if re.match(regex, value) is False and value != "":
            SERVER_LOG.critical(
                f'Попытка запуска сервера с указанием неподходящего хоста '
                f'{value}. Допустимы адреса IPV4.')
            sys.exit(1)
        instance.__dict__[self.name] = value

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        return instance.__dict__[self.name]



import binascii
import hashlib
import sys

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QMainWindow, QAction, qApp, QApplication, QLabel, QTableView, QDialog, QPushButton, \
    QLineEdit, QFileDialog, QMessageBox, QComboBox


# Класс основного окна
class MainWindow(QMainWindow):
    '''Класс - основное окно сервера.'''

    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        # Кнопка выхода
        exitAction = QAction('Выход', self)
        exitAction.setShortcut('Ctrl+Q')
        exitAction.triggered.connect(qApp.quit)

        # Кнопка обновить список клиентов
        self.refresh_button = QAction('Обновить список', self)

        # Кнопка настроек сервера
        self.config_btn = QAction('Настройки сервера', self)

        # Кнопка вывести историю сообщений
        self.show_history_button = QAction('История клиентов', self)

        # Кнопка регистрации пользователя
        self.register_btn = QAction('Регистрация пользователя', self)

        # Кнопка удаления пользователя
        self.remove_btn = QAction('Удаление пользователя', self)

        # Статусбар
        # dock widget
        self.statusBar()

        # Тулбар
        self.toolbar = self.addToolBar('MainBar')
        self.toolbar.addAction(exitAction)
        self.toolbar.addAction(self.refresh_button)
        self.toolbar.addAction(self.show_history_button)
        self.toolbar.addAction(self.config_btn)
        self.toolbar.addAction(self.register_btn)
        self.toolbar.addAction(self.remove_btn)

        # Настройки геометрии основного окна
        # Поскольку работать с динамическими размерами мы не умеем, и мало времени на изучение, размер окна фиксирован.
        self.setFixedSize(800, 600)
        self.setWindowTitle('Messaging Server alpha release')

        # Надпись о том, что ниже список подключённых клиентов
        self.label = QLabel('Список подключённых клиентов:', self)
        self.label.setFixedSize(240, 15)
        self.label.move(10, 25)

        # Окно со списком подключённых клиентов.
        self.active_clients_table = QTableView(self)
        self.active_clients_table.move(10, 45)
        self.active_clients_table.setFixedSize(780, 400)

        # Последним параметром отображаем окно.
        self.show()


class HistoryWindow(QDialog):
    """
    # Класс окна с историей пользователей
    """
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        # Настройки окна:
        self.setWindowTitle('Статистика клиентов')
        self.setFixedSize(600, 700)
        self.setAttribute(Qt.WA_DeleteOnClose)

        # Кнапка закрытия окна
        self.close_button = QPushButton('Закрыть', self)
        self.close_button.move(250, 650)
        self.close_button.clicked.connect(self.close)

        # Лист с собственно историей
        self.history_table = QTableView(self)
        self.history_table.move(10, 10)
        self.history_table.setFixedSize(580, 620)

        self.show()


# Класс окна настроек
class ConfigWindow(QDialog):
    """
    Класс окна конфигурации
    """
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        """# Настройки окна"""
        self.setFixedSize(365, 260)
        self.setWindowTitle('Настройки сервера')

        # Надпись о файле базы данных:
        self.db_path_label = QLabel('Путь до файла базы данных: ', self)
        self.db_path_label.move(10, 10)
        self.db_path_label.setFixedSize(240, 15)

        # Строка с путём базы
        self.db_path = QLineEdit(self)
        self.db_path.setFixedSize(250, 20)
        self.db_path.move(10, 30)
        self.db_path.setReadOnly(True)

        # Кнопка выбора пути.
        self.db_path_select = QPushButton('Обзор...', self)
        self.db_path_select.move(275, 28)

        def open_file_dialog():
            """
            # Функция обработчик открытия окна выбора папки
            @return:
            """
            global dialog
            dialog = QFileDialog.getOpenFileName(self)[0]
            # path = dialog.getExistingDirectory()
            path = dialog
            # path = path.replace('/', '\\')
            print(path.split('/')[0:-1])
            self.db_path.insert('/'.join(path.split('/')[0:-1]))
            self.db_file.insert(path.split('/')[-1])

        self.db_path_select.clicked.connect(open_file_dialog)

        # Метка с именем поля файла базы данных
        self.db_file_label = QLabel('Имя файла базы данных: ', self)
        self.db_file_label.move(10, 68)
        self.db_file_label.setFixedSize(180, 15)

        # Поле для ввода имени файла
        self.db_file = QLineEdit(self)
        self.db_file.move(200, 66)
        self.db_file.setFixedSize(150, 20)

        # Метка с номером порта
        self.port_label = QLabel('Номер порта для соединений:', self)
        self.port_label.move(10, 108)
        self.port_label.setFixedSize(180, 15)

        # Поле для ввода номера порта
        self.port = QLineEdit(self)
        self.port.move(200, 108)
        self.port.setFixedSize(150, 20)

        # Метка с адресом для соединений
        self.ip_label = QLabel('С какого IP принимаем соединения:', self)
        self.ip_label.move(10, 148)
        self.ip_label.setFixedSize(180, 15)

        # Метка с напоминанием о пустом поле.
        self.ip_label_note = QLabel(' оставьте это поле пустым, чтобы\n принимать соединения с любых адресов.', self)
        self.ip_label_note.move(10, 168)
        self.ip_label_note.setFixedSize(500, 30)

        # Поле для ввода ip
        self.ip = QLineEdit(self)
        self.ip.move(200, 148)
        self.ip.setFixedSize(150, 20)

        # Кнопка сохранения настроек
        self.save_btn = QPushButton('Сохранить', self)
        self.save_btn.move(190, 220)

        # Кнапка закрытия окна
        self.close_button = QPushButton('Закрыть', self)
        self.close_button.move(275, 220)
        self.close_button.clicked.connect(self.close)

        self.show()


class RegisterUser(QDialog):
    '''Класс диалог регистрации пользователя на сервере.'''

    def __init__(self, database, server):
        super().__init__()

        self.database = database
        self.server = server

        self.setWindowTitle('Регистрация')
        self.setFixedSize(175, 183)
        self.setModal(True)
        self.setAttribute(Qt.WA_DeleteOnClose)

        self.label_username = QLabel('Введите имя пользователя:', self)
        self.label_username.move(10, 10)
        self.label_username.setFixedSize(150, 15)

        self.client_name = QLineEdit(self)
        self.client_name.setFixedSize(154, 20)
        self.client_name.move(10, 30)

        self.label_passwd = QLabel('Введите пароль:', self)
        self.label_passwd.move(10, 55)
        self.label_passwd.setFixedSize(150, 15)

        self.client_passwd = QLineEdit(self)
        self.client_passwd.setFixedSize(154, 20)
        self.client_passwd.move(10, 75)
        self.client_passwd.setEchoMode(QLineEdit.Password)
        self.label_conf = QLabel('Введите подтверждение:', self)
        self.label_conf.move(10, 100)
        self.label_conf.setFixedSize(150, 15)

        self.client_conf = QLineEdit(self)
        self.client_conf.setFixedSize(154, 20)
        self.client_conf.move(10, 120)
        self.client_conf.setEchoMode(QLineEdit.Password)

        self.btn_ok = QPushButton('Сохранить', self)
        self.btn_ok.move(10, 150)
        self.btn_ok.clicked.connect(self.save_data)

        self.btn_cancel = QPushButton('Выход', self)
        self.btn_cancel.move(90, 150)
        self.btn_cancel.clicked.connect(self.close)

        self.messages = QMessageBox()

        self.show()

    def save_data(self):
        '''
        Метод проверки правильности ввода и сохранения в базу нового пользователя.
        '''
        if not self.client_name.text():
            self.messages.critical(
                self, 'Ошибка', 'Не указано имя пользователя.')
            return
        elif self.client_passwd.text() != self.client_conf.text():
            self.messages.critical(
                self, 'Ошибка', 'Введённые пароли не совпадают.')
            return
        elif self.database.check_user(self.client_name.text()):
            self.messages.critical(
                self, 'Ошибка', 'Пользователь уже существует.')
            return
        else:
            # Генерируем хэш пароля, в качестве соли будем использовать логин в
            # нижнем регистре.
            passwd_bytes = self.client_passwd.text().encode('utf-8')
            salt = self.client_name.text().lower().encode('utf-8')
            passwd_hash = hashlib.pbkdf2_hmac(
                'sha512', passwd_bytes, salt, 10000)
            self.database.add_user(
                self.client_name.text(),
                binascii.hexlify(passwd_hash))
            self.messages.information(
                self, 'Успех', 'Пользователь успешно зарегистрирован.')
            # Рассылаем клиентам сообщение о необходимости обновить справичники
            self.server.service_update_lists()
            self.close()


class DelUserDialog(QDialog):
    '''
    Класс - диалог выбора контакта для удаления.
    '''

    def __init__(self, database, server):
        super().__init__()
        self.database = database
        self.server = server

        self.setFixedSize(350, 120)
        self.setWindowTitle('Удаление пользователя')
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setModal(True)

        self.selector_label = QLabel(
            'Выберите пользователя для удаления:', self)
        self.selector_label.setFixedSize(200, 20)
        self.selector_label.move(10, 0)

        self.selector = QComboBox(self)
        self.selector.setFixedSize(200, 20)
        self.selector.move(10, 30)

        self.btn_ok = QPushButton('Удалить', self)
        self.btn_ok.setFixedSize(100, 30)
        self.btn_ok.move(230, 20)
        self.btn_ok.clicked.connect(self.remove_user)

        self.btn_cancel = QPushButton('Отмена', self)
        self.btn_cancel.setFixedSize(100, 30)
        self.btn_cancel.move(230, 60)
        self.btn_cancel.clicked.connect(self.close)

        self.all_users_fill()

    def all_users_fill(self):
        '''Метод заполняющий список пользователей.'''
        self.selector.addItems([item[0]
                                for item in self.database.users_list()])

    def remove_user(self):
        '''Метод - обработчик удаления пользователя.'''
        self.database.remove_user(self.selector.currentText())
        if self.selector.currentText() in self.server.names:
            sock = self.server.names[self.selector.currentText()]
            del self.server.names[self.selector.currentText()]
            self.server.remove_client(sock)
        # Рассылаем клиентам сообщение о необходимости обновить справочники
        self.server.service_update_lists()
        self.close()


if __name__ == '__main__':
    app = QApplication(sys.argv)
    ex = MainWindow()
    ex.statusBar().showMessage('Test Statusbar Message')
    test_list = QStandardItemModel(ex)
    test_list.setHorizontalHeaderLabels(['Имя Клиента', 'IP Адрес', 'Порт', 'Время подключения'])
    test_list.appendRow([QStandardItem('1'), QStandardItem('2'), QStandardItem('3')])
    test_list.appendRow([QStandardItem('4'), QStandardItem('5'), QStandardItem('6')])
    ex.active_clients_table.setModel(test_list)
    ex.active_clients_table.resizeColumnsToContents()
    message = QMessageBox
    # dial = ConfigWindow()

    app.exec_()



import functools
import inspect
import logging
import sys
import traceback

if sys.argv[0].find('client') == -1:
    # если не клиент, то сервер!
    LOGGER = logging.getLogger('app.server')
else:
    # ну, раз не сервер, то клиент
    LOGGER = logging.getLogger('app.client')


# Реализация в виде класса для функции
class Log:
    """Класс-декоратор"""

    def __call__(self, cls):
        def decorate(func_log):
            def log_saver(*args, **kwargs):
                """Обертка"""
                LOGGER.debug(
                    f'Была вызвана функция {func_log.__name__} c параметрами args={args}, kwargs={kwargs}. '
                    f'Вызов из модуля {func_log.__module__}. Вызов из'
                    f' объекта {traceback.format_stack()[0].strip().split()[-1]}.'
                    f'Вызов из функции {inspect.stack()[1][3]}', stacklevel=2)
                return_value = func_log(*args, **kwargs)
                return return_value

            return log_saver

        cls.__init__ = decorate(cls.__init__)
        cls.process_client_message = decorate(cls.process_client_message)
        return cls


# Реализация в виде функции
def log(func_log):
    """Функция-декоратор"""

    @functools.wraps(func_log)
    def log_saver(*args, **kwargs):
        """Обертка"""
        LOGGER.debug(
            f'log-func: Была вызвана функция {func_log.__name__} c параметрами args={args}, kwargs={kwargs}. '
            f'Вызов из модуля {func_log.__module__}. Вызов из'
            f' объекта {traceback.format_stack()[0].strip().split()[-1]}.'
            f'Вызов из функции {inspect.stack()[1][3]}', stacklevel=2)
        return_value = func_log(*args, **kwargs)
        return return_value

    return log_saver



# Файл констант

DEFAULT_PORT = 8080
DEFAULT_IP_ADDRESS = '127.0.0.1'
MAX_CONNECTIONS = 10
MAX_PACKET_LENGTH = 4096
ENCODING_FIELD = 'encoding'
ENCODING = 'utf-8'

ACTION = 'action'
EXIT = 'exit'
TIME = 'time'
USER = 'user'
ACCOUNT_NAME = 'account_name'

PRESENCE = 'presence'
RESPONSE = 'response'
ERROR = 'error'
DATA = 'data'

AUTHENTICATE = 'authenticate'
QUIT = 'quit'
PROBE = 'probe'
MESSAGE = 'msg'
JOIN_GROUP = 'join'
LEAVE_GROUP = 'leave'
STATUS = 'status'
LOGGING_LEVEL = 'DEBUG'
USERS_REQUEST = 'users_request'

MESSAGE_TEXT = 'message_text'
SENDER = 'sender'
TO = 'to'
FROM = 'from'
GET_CONTACTS = 'get_contacts'
ADD_CONTACT = 'add_contact'
DEL_CONTACT = 'del_contact'
LIST_INFO = 'list_info'
GET_LIST = "get_list"
PUBLIC_KEY = "public_key"
PUBLIC_KEY_REQUEST = "public_key_request"

# Словари - ответы:
# 200
RESPONSE_200 = {RESPONSE: 200}
# 400
RESPONSE_400 = {
    RESPONSE: 400,
    ERROR: None
}

RESPONSE_205 = {RESPONSE: 205}
RESPONSE_511 = {
    RESPONSE: 511,
    DATA: ''}



import logging
import argparse
import configparser
import select
import socket
import sys
import os
import threading
import hmac
import binascii

from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QApplication, QMessageBox

sys.path.append('../')
from common.variables import *
from common.utils import Connector, get_message, send_message
from db.db_server import ServerDB
from common.metaclasses import ServerVerifier
from common.descriptors import Port, Host
from server.gui_server import MainWindow, HistoryWindow, ConfigWindow, RegisterUser, DelUserDialog


# Флаг что был подключён новый пользователь, нужен чтобы не мучать BD
# постоянными запросами на обновление
new_connection = False
conflag_lock = threading.Lock()


class Server(threading.Thread, metaclass=ServerVerifier):
    listen_port = Port()
    listen_address = Host()

    def argsParser(self, arguments=None):
        """Парсер аргументов коммандной строки"""
        if arguments is None:
            arguments = []

        argument_parser = argparse.ArgumentParser()
        argument_parser.add_argument('-a', default='', nargs='?')
        argument_parser.add_argument(
            '-p', default=DEFAULT_PORT, type=int, nargs='?')
        namespace = argument_parser.parse_args(sys.argv[1:])
        self.listen_address = namespace.a
        self.listen_port = namespace.p

    def __init__(self, db, arguments=None):
        if arguments is None:
            arguments = []

        self.SERVER_LOG = logging.getLogger('app.server')
        # список клиентов , очередь сообщений
        self.clients = []
        self.messages = []
        self.names = dict()
        self.database = db
        self.argsParser()
        self.active_users = []
        self.history_users = self.database.message_history()

        super().__init__()

    def process_client_message(self, message, client):
        """
        Обработка входящего сообщения
        @param message: сообщение
        @param client: от кого пришло
        @return:
        """
        global new_connection
        # Проверяем, если это сообщение о присутствии, принимаем и отвечаем,
        # если успех
        if ACTION in message and message[ACTION] == PRESENCE and USER in message:
            # Если сообщение о присутствии то вызываем функцию авторизации.
            self.autorize_user(message, client)

        # Если это сообщение, то добавляем его в очередь сообщений. Ответ не
        # требуется.
        elif ACTION in message and message[ACTION] == MESSAGE and MESSAGE_TEXT in message and \
                TO in message and FROM in message:
            self.database.process_message(message[FROM], message[TO])
            self.history_users = self.database.message_history()
            self.messages.append(message)
            return

        # Запрос контакт листа пользователей с сервера:
        elif ACTION in message and message[ACTION] == GET_CONTACTS and USER in message:
            query = self.database.get_contacts(message[USER])
            response_contacts_list = []
            for user in query:
                response_contacts_list.append(user)
            response = {
                RESPONSE: 200,
                LIST_INFO: response_contacts_list
            }
            Connector(client).send_message(response)

        elif ACTION in message and message[ACTION] == ADD_CONTACT \
                and ACCOUNT_NAME in message and USER in message:
            self.database.add_contact(message[ACCOUNT_NAME], message[USER])
            Connector(client).send_message(RESPONSE_200)

        elif ACTION in message and message[ACTION] == DEL_CONTACT and ACCOUNT_NAME in message \
                and USER in message:
            self.database.remove_contact(message[ACCOUNT_NAME], message[USER])
            Connector(client).send_message(RESPONSE_200)

        # Если клиент выходит
        elif ACTION in message and message[ACTION] == EXIT and ACCOUNT_NAME in message:
            self.database.user_logout(message[ACCOUNT_NAME])
            self.active_users = self.database.active_users_list()
            self.clients.remove(self.names[message[ACCOUNT_NAME]])
            self.names[message[ACCOUNT_NAME]].close()
            del self.names[message[ACCOUNT_NAME]]
            with conflag_lock:
                new_connection = True
            return

        elif ACTION in message and message[ACTION] == GET_LIST:
            query = self.database.users_list()
            print(f'We get message: {message}')
            response_temp = []
            for user in query:
                response_temp.append(user[0])
            response = {
                RESPONSE: 200,
                LIST_INFO: response_temp
            }
            Connector(client).send_message(response)

        elif ACTION in message and message[ACTION] == "GET_HISTORY" and "history_name" in message:
            query = self.database.login_history(message["history_name"])
            response_temp = []
            for user in query:
                response_temp.append(
                    (user[0], user[1].strftime("%m/%d/%Y, %H:%M:%S")))
            response = {
                ACTION: "GET_HISTORY",
                MESSAGE_TEXT: response_temp
            }
            Connector(client).send_message(response)
        elif ACTION in message and message[ACTION] == "GET_CONNLIST":
            query = self.database.active_users_list()
            response_temp = []
            for user in query:
                response_temp.append(user[0])
            print(response_temp)
            response = {
                ACTION: "GET_CONNLIST",
                MESSAGE_TEXT: response_temp
            }
            Connector(client).send_message(response)
        elif ACTION in message and message[ACTION] == PUBLIC_KEY_REQUEST and ACCOUNT_NAME in message:
            response = RESPONSE_511
            response[DATA] = self.database.get_pubkey(message[ACCOUNT_NAME])
            # может быть, что ключа ещё нет (пользователь никогда не логинился,
            # тогда шлём 400)
            if response[DATA]:
                try:
                    send_message(client, response)
                except OSError:
                    self.remove_client(client)
            else:
                response = RESPONSE_400
                response[ERROR] = 'Нет публичного ключа для данного пользователя'
                try:
                    send_message(client, response)
                except OSError:
                    self.remove_client(client)

        # Иначе отдаём Bad request
        else:
            response = RESPONSE_400
            response[ERROR] = 'Запрос некорректен.'
            Connector(client).send_message(response)
            return

    def process_message(self, message, listen_socks):
        """
        Функция адресной отправки сообщения определённому клиенту. Принимает словарь сообщение,
        список зарегистрированых пользователей и слушающие сокеты. Ничего не возвращает.
        """
        if message[TO] in self.names and self.names[message[TO]] in listen_socks:
            Connector(self.names[message[TO]]).send_message(message)
            self.SERVER_LOG.info(
                f'Отправлено сообщение пользователю {message[TO]} '
                f'от пользователя {message[FROM]}.')
        elif message[TO] in self.names.keys() and self.names[message[TO]] not in listen_socks:
            raise ConnectionError
        else:
            self.SERVER_LOG.error(
                f'Пользователь {message[TO]} не зарегистрирован на сервере, '
                f'отправка сообщения невозможна.')

    def init_socket(self):
        '''Метод инициализатор сокета.'''
        self.SERVER_LOG.debug(
            (f'Запущен сервер, порт для подключений: {self.listen_port}, '
             f'адрес с которого принимаются подключения: {self.listen_address}. '
             f'Если адрес не указан, принимаются соединения с любых адресов.'))
        transport = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        transport.bind((self.listen_address, self.listen_port))
        transport.settimeout(0.5)

        # Начинаем слушать сокет.
        self.SERV_SOCK = transport
        self.SERV_SOCK.listen()

    def run(self):
        """
        Функция потока
        @return:
        """
        self.init_socket()
        while True:
            try:
                CLIENT_SOCK, CLIENT_ADDR = self.SERV_SOCK.accept()
            except OSError:
                pass
            else:
                self.SERVER_LOG.debug(
                    f'Установлено соединение с ПК по адресу: {CLIENT_ADDR}')
                self.clients.append(CLIENT_SOCK)

            # Начинаем работать с несколькими клиентами и информацией от них
            received_data_lst = []
            send_data_lst = []
            error_lst = []
            # Словарь, содержащий имена пользователей и соответствующие им сокеты.
            # Проверяем на наличие ждущих клиентов
            try:
                if self.clients:
                    received_data_lst, send_data_lst, error_lst = select.select(
                        self.clients, self.clients, [], 0)
            except OSError:
                pass
            # принимаем сообщения и если там есть сообщения,
            # кладём в словарь, если ошибка, исключаем клиента.
            if received_data_lst:
                print(f"received_data_lst in server: {received_data_lst}")
                for client_with_message in received_data_lst:
                    print(
                        f"client_with_message in received_data_lst: {client_with_message}")
                    try:
                        message_from_client = get_message(client_with_message)
                        print(f"message_from_client: {message_from_client}")
                        self.process_client_message(
                            message_from_client, client_with_message)
                        self.SERVER_LOG.info(
                            f'Список принятых сообщений: {self.messages}')
                    except BaseException:
                        self.SERVER_LOG.info(
                            f'Клиент {client_with_message.getpeername()} '
                            f'отключился от сервера.')
                        self.clients.remove(client_with_message)
            # Если есть сообщения для отправки и ожидающие клиенты, отправляем
            # им сообщение.
            for message in self.messages:
                try:
                    self.process_message(message, send_data_lst)
                except Exception:
                    self.SERVER_LOG.info(
                        f'Клиент {message[TO]} отключился от сервера.')
                    self.clients.remove(self.names[message[TO]])
                    del self.names[message[TO]]
            self.messages.clear()

    # GUI - Создание таблицы QModel, для отображения в окне программы.
    def gui_create_model(self):
        list = QStandardItemModel()
        list.setHorizontalHeaderLabels(
            ['Имя Клиента', 'IP Адрес', 'Порт', 'Время подключения'])
        for row in self.active_users:
            user, ip, port, time = row
            user = QStandardItem(user)
            user.setEditable(False)
            ip = QStandardItem(ip)
            ip.setEditable(False)
            port = QStandardItem(str(port))
            port.setEditable(False)
            time = QStandardItem(str(time.replace(microsecond=0)))
            time.setEditable(False)
            list.appendRow([user, ip, port, time])
        return list

    # GUI - Функция реализующая заполнение таблицы историей сообщений.
    def gui_create_stat_model(self):
        list = QStandardItemModel()
        list.setHorizontalHeaderLabels(
            ['Имя Клиента', 'Последний раз входил', 'Сообщений отправлено', 'Сообщений получено'])
        for row in self.history_users:
            user, last_seen, sent, recvd = row
            user = QStandardItem(user)
            user.setEditable(False)
            last_seen = QStandardItem(str(last_seen.replace(microsecond=0)))
            last_seen.setEditable(False)
            sent = QStandardItem(str(sent))
            sent.setEditable(False)
            recvd = QStandardItem(str(recvd))
            recvd.setEditable(False)
            list.appendRow([user, last_seen, sent, recvd])
        return list

    def autorize_user(self, message, sock):
        '''Метод реализующий авторизцию пользователей.'''
        global new_connection
        # Если имя пользователя уже занято то возвращаем 400
        self.SERVER_LOG.debug(f'Start auth process for {message[USER]}')
        print(f'Start auth process for {message[USER]}')
        if message[USER][ACCOUNT_NAME] in self.names.keys():
            response = RESPONSE_400
            response[ERROR] = 'Имя пользователя уже занято.'
            try:
                self.SERVER_LOG.debug(f'Username busy, sending {response}')
                send_message(sock, response)
            except OSError:
                self.SERVER_LOG.debug('OS Error')
                pass
            self.clients.remove(sock)
            sock.close()
        # Проверяем что пользователь зарегистрирован на сервере.
        elif not self.database.check_user(message[USER][ACCOUNT_NAME]):
            response = RESPONSE_400
            response[ERROR] = 'Пользователь не зарегистрирован.'
            try:
                self.SERVER_LOG.debug(f'Unknown username, sending {response}')
                send_message(sock, response)
            except OSError:
                pass
            self.clients.remove(sock)
            sock.close()
        else:
            self.SERVER_LOG.debug('Correct username, starting passwd check.')
            # Иначе отвечаем 511 и проводим процедуру авторизации
            # Словарь - заготовка
            message_auth = RESPONSE_511
            print(f"message_auth = RESPONSE_511: {message_auth}")
            # Набор байтов в hex представлении
            random_str = binascii.hexlify(os.urandom(64))
            print(f"random_str: {random_str}")
            # В словарь байты нельзя, декодируем (json.dumps -> TypeError)
            message_auth[DATA] = random_str.decode('ascii')
            print(f"message_auth = RESPONSE_511: {message_auth}")
            # Создаём хэш пароля и связки с рандомной строкой, сохраняем
            # серверную версию ключа
            hash = hmac.new(
                self.database.get_hash(
                    message[USER][ACCOUNT_NAME]),
                random_str,
                'MD5')
            digest = hash.digest()
            self.SERVER_LOG.debug(f'Auth message = {message_auth}')
            print(f'Auth message = {message_auth}')
            try:
                # Обмен с клиентом
                send_message(sock, message_auth)
                ans = get_message(sock)
            except OSError as err:
                self.SERVER_LOG.debug('Error in auth, data:', exc_info=err)
                sock.close()
                return
            client_digest = binascii.a2b_base64(ans[DATA])
            # Если ответ клиента корректный, то сохраняем его в список
            # пользователей.
            if RESPONSE in ans and ans[RESPONSE] == 511 and hmac.compare_digest(
                    digest, client_digest):
                self.names[message[USER][ACCOUNT_NAME]] = sock
                client_ip, client_port = sock.getpeername()
                try:
                    send_message(sock, RESPONSE_200)
                except OSError:
                    self.remove_client(message[USER][ACCOUNT_NAME])
                # добавляем пользователя в список активных и если у него изменился открытый ключ
                # сохраняем новый
                self.database.user_login(
                    message[USER][ACCOUNT_NAME],
                    client_ip,
                    client_port,
                    message[USER][PUBLIC_KEY])
                self.active_users = self.database.active_users_list()
                new_connection = True
            else:
                response = RESPONSE_400
                response[ERROR] = 'Неверный пароль.'
                try:
                    send_message(sock, response)
                except OSError:
                    pass
                self.clients.remove(sock)
                sock.close()

    def service_update_lists(self):
        '''Метод реализующий отправки сервисного сообщения 205 клиентам.'''
        for client in self.clients:
            try:
                Connector(client).send_message(RESPONSE_205)
            except OSError:
                self.clients.remove(client)


def main():
    config = configparser.ConfigParser()

    dir_path = os.getcwd()
    config.read(f"{dir_path}/{'server.ini'}")

    database = ServerDB(
        os.path.join(
            config['SETTINGS']['Database_path'],
            config['SETTINGS']['Database_file']))

    server = Server(database, f"- a {config['SETTINGS']['Default_port']}"
                              f" - p {config['SETTINGS']['Listen_Address']}")

    server.daemon = True
    server.start()

    server_app = QApplication(sys.argv)
    main_window = MainWindow()

    main_window.statusBar().showMessage('Server Working')

    main_window.active_clients_table.setModel(server.gui_create_model())
    main_window.active_clients_table.resizeColumnsToContents()
    main_window.active_clients_table.resizeRowsToContents()

    def list_update():
        global new_connection
        if new_connection:
            main_window.active_clients_table.setModel(
                server.gui_create_model())
            main_window.active_clients_table.resizeColumnsToContents()
            main_window.active_clients_table.resizeRowsToContents()
            with conflag_lock:
                new_connection = False

    # Функция создающяя окно со статистикой клиентов
    def show_statistics():
        global stat_window
        stat_window = HistoryWindow()
        stat_window.history_table.setModel(server.gui_create_stat_model())
        stat_window.history_table.resizeColumnsToContents()
        stat_window.history_table.resizeRowsToContents()
        stat_window.show()

    # Функция создающяя окно с настройками сервера.
    def server_config():
        global config_window
        # Создаём окно и заносим в него текущие параметры
        config_window = ConfigWindow()
        config_window.db_path.insert(config['SETTINGS']['Database_path'])
        config_window.db_file.insert(config['SETTINGS']['Database_file'])
        config_window.port.insert(config['SETTINGS']['Default_port'])
        config_window.ip.insert(config['SETTINGS']['Listen_Address'])
        config_window.save_btn.clicked.connect(save_server_config)

    # Функция сохранения настроек
    def save_server_config():
        global config_window
        message = QMessageBox()
        config['SETTINGS']['Database_path'] = config_window.db_path.text()
        config['SETTINGS']['Database_file'] = config_window.db_file.text()
        try:
            port = int(config_window.port.text())
        except ValueError:
            message.warning(config_window, 'Ошибка', 'Порт должен быть числом')
        else:
            config['SETTINGS']['Listen_Address'] = config_window.ip.text()
            if 1023 < port < 65536:
                config['SETTINGS']['Default_port'] = str(port)
                print(port)
                with open('server.ini', 'w') as conf:
                    config.write(conf)
                    message.information(
                        config_window, 'OK', 'Настройки успешно сохранены!')
            else:
                message.warning(
                    config_window,
                    'Ошибка',
                    'Порт должен быть от 1024 до 65536')

    def reg_user():
        '''Метод создающий окно регистрации пользователя.'''
        global reg_window
        reg_window = RegisterUser(database, server)
        reg_window.show()

    def rem_user():
        '''Метод создающий окно удаления пользователя.'''
        global rem_window
        rem_window = DelUserDialog(database, server)
        rem_window.show()

    # Таймер, обновляющий список клиентов 1 раз в секунду
    timer = QTimer()
    timer.timeout.connect(list_update)
    timer.start(1000)

    # Связываем кнопки с процедурами
    main_window.refresh_button.triggered.connect(list_update)
    main_window.show_history_button.triggered.connect(show_statistics)
    main_window.config_btn.triggered.connect(server_config)
    main_window.register_btn.triggered.connect(reg_user)
    main_window.remove_btn.triggered.connect(rem_user)

    # Запускаем GUI
    server_app.exec_()


if __name__ == '__main__':
    main()



import json

from .decos import log
from .variables import *


class Connector:
    def __init__(self, sock):
        self.sock = sock

    def get_message(self):
        encoded_response = self.sock.recv(MAX_PACKET_LENGTH)
        if isinstance(encoded_response, bytes):
            json_response = encoded_response.decode(ENCODING)
            response = json.loads(json_response)
            if isinstance(response, dict):
                return response
            raise ValueError
        raise ValueError

    def send_message(self, message):
        js_message = json.dumps(message, ensure_ascii=False)
        encoded_message = js_message.encode(ENCODING)
        self.sock.send(encoded_message)


# if __name__ == "__main__":
#     message = {
#         "test": [('client_2', '192.168.1.5', 7777, datetime.datetime(2021, 10, 28, 15, 5, 49, 716620)), ('client_3', '192.168.1.6', 6666, datetime.datetime(2021, 10, 28, 15, 5, 49, 763465)), ('client_4', '192.168.1.7', 5555, datetime.datetime(2021, 10, 28, 15, 5, 49, 810328))]
#     }
#     js_message = json.dumps(message, ensure_ascii=False)

@log
def get_message(client):
    encoded_response = client.recv(MAX_PACKET_LENGTH)
    json_response = encoded_response.decode(ENCODING)
    response = json.loads(json_response)
    if isinstance(response, dict):
        return response
    else:
        raise TypeError


# Утилита кодирования и отправки сообщения
# принимает словарь и отправляет его
@log
def send_message(sock, message):
    js_message = json.dumps(message, ensure_ascii=False)
    encoded_message = js_message.encode(ENCODING)
    sock.send(encoded_message)



import dis
import logging

SERVER_LOG = logging.getLogger('app.server')
CLIENT_LOG = logging.getLogger('app.client')


class ServerVerifier(type):
    def __init__(self, clsname, bases, clsdict):
        """
        clsname - экземпляр класса
        bases - кортеж предков
        clsdict - словарь атрибутов и методов
        """
        methods = []  # Пустой список методов
        attributes = []  # Пустой список атрибутов

        for function in clsdict:
            try:
                returns = dis.get_instructions(clsdict[function])
            except TypeError:  # exception if not a function in dis
                pass
            else:  # if function
                for i in returns:
                    if i.opname == 'LOAD_GLOBAL':
                        if i.argval not in methods:  # заполняем список методами, использующимися в функциях класса
                            methods.append(i.argval)
                    elif i.opname == 'LOAD_ATTR':
                        if i.argval not in attributes:  # заполняем список атрибутами, использующимися в функциях класса
                            attributes.append(i.argval)
        # Если обнаружено использование недопустимого метода connect, бросаем
        # исключение:
        SERVER_LOG.debug(
            f"Нашли методы при инициализации фабрикой классов: {methods}")
        SERVER_LOG.debug(
            f"Нашли атрибуты при инициализации фабрикой классов: {attributes}")

        if 'connect' in methods:
            raise TypeError(
                'Использование метода connect недопустимо в серверном классе')
        # Если сокет не инициализировался константами SOCK_STREAM(TCP)
        # AF_INET(IPv4), тоже исключение.
        if not ('SOCK_STREAM' in attributes and 'AF_INET' in attributes):
            raise TypeError('Некорректная инициализация сокета.')

        super().__init__(clsname, bases, clsdict)


class ClientVerifier(type):
    def __init__(self, clsname, bases, clsdict):
        methods = []  # Пустой список методов
        attributes = []  # Пустой список атрибутов
        print(clsdict)
        for function in clsdict:
            try:
                returns = dis.get_instructions(clsdict[function])
            except TypeError:
                pass
            else:
                for i in returns:
                    if i.opname == 'LOAD_GLOBAL':
                        if i.argval not in methods:
                            methods.append(i.argval)
                    elif i.opname == 'LOAD_ATTR':
                        if i.argval not in attributes:  # заполняем список атрибутами, использующимися в функциях класса
                            attributes.append(i.argval)
        CLIENT_LOG.debug(
            f"Нашли методы при инициализации фабрикой классов: {methods}")
        CLIENT_LOG.debug(
            f"Нашли атрибуты при инициализации фабрикой классов: {attributes}")

        # Если обнаружено использование недопустимого метода accept, listen,
        # socket бросаем исключение:
        for command in ('accept', 'listen', 'socket'):
            if command in methods:
                raise TypeError(
                    'В классе обнаружено использование запрещённого метода')
        # Наличие поле connector, который является объектом с методами работы с
        # сокетами считаем корректным использованием сокетов
        if 'connector' in attributes:
            pass
        else:
            raise TypeError('Отсутствует поле, работающее с сокетами.')
        super().__init__(clsname, bases, clsdict)
